package funktiopiirturi;

import java.util.Hashtable;
import java.util.List;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Set;
import java.util.TreeSet;
import java.util.Stack;
import math.*;

/**
 * FuncXtoY esitt muotoa y = f(x) olevaa funktiota, jonka arvon voi laskea antamalla x.
 *
 */
public class FuncXtoY {
	// Staattiset muuttujat
	private static Set<String> functions;		// kytettviss olevat funktiot (sin, ...)
	private static Hashtable<String, Expression> symbols;	// symbolitaulu
	private static Constant CONST_PI, CONST_E;				// vakioita
	private static Variable varX;			// x-muuttuja, jonka perusteella arvo lasketaan
	
	// Oliokohtaiset muuttujat
	private Expression rightSide;	// funktion oikean puolen lauseke
	
	/**
	 * Alustetaa staattiset muuttujat/vakiot.
	 */
	static {
		// listataan kytettviss olevat funktiot
		functions = new TreeSet<String>();
		functions.add("sqrt");
		functions.add("cbrt");
		functions.add("ln");
		functions.add("lg");
		functions.add("exp");
		functions.add("abs");
		functions.add("sin");
		functions.add("cos");
		functions.add("tan");
		functions.add("asin");
		functions.add("acos");
		functions.add("atan");
		functions.add("sinh");
		functions.add("cosh");
		functions.add("tanh");
		
		
		// luodaan symbolitaulu ja listn siihen muutama vakio ja muuttuja
		symbols = new Hashtable<String, Expression>();
		
		CONST_PI = new Constant(Math.PI);
		CONST_E = new Constant(Math.E);
		
		symbols.put("pi", CONST_PI);
		symbols.put("e", CONST_E);
		
		varX = new Variable();
		symbols.put("x", varX);
	}
	
	/**
	 * Pilkkoo parametrina annetun matemaattisen lausekkeen alkeisosasiksi (operaatioihin,
	 * funktioihin, muutujiin, vakioihin, sulkuihin). Palauttaa String-listan, jossa
	 * alkeisosaset ovat erikseen.
	 * @param s		pilkottava lauseke/funktio
	 * @return		lista, jossa lauseke alkeisosasina
	 */
	private static List<String> Split(String s) {
		s = s.replaceAll("\\s+", " ");	//  korvataan kaikki vlit yhdell vlilynnill
		List<String> atoms = new LinkedList<String>();
		int i = 0;	// tmnhetkinen merkki
		int b = 0;	// edellisen erotinmerkin jlkeinen merkki (lohkon alku)
		while( i < s.length() ) {
			// etsitn seuraava erotinmerkki
			while( i < s.length() && "()+-*/^# ".indexOf(s.charAt(i)) == -1 ) {
				if( s.charAt(i) == 'E' )	// exp.muod. annettu luku => skipataan mahd +/-
					if( i+1 < s.length() && "+-".indexOf(s.charAt(i+1)) >= 0 )
						++i;
				++i;
			}
			// jos edeltv lohko ei ollut erotinmerkki, laitetaan se talteen
			if( i > b )
				atoms.add(s.substring(b, i));
			// laitetaan erotinmerkki talteen, ja siirrytn seuraavaan kohtaan
			b = i + 1;
			if( i < s.length() && s.charAt(i) != ' ' )
				atoms.add(s.substring(i, b));
			++i;
		}
		return atoms;
	}
	
	/**
	 * Rakentaa parametrina annetusta matemaattisesta alkiolistasta rekursiivisen puumaisen
	 * lausekerakenteen. Mahdottoman kohdan sattuessa heitt asiaankuluvan poikkeuksen.
	 * @param atoms	alkiolista, joka sislt matemaattisen lausekkeen alkeisosaset oikeassa
	 * 				jrjestyksess. Alkiot ovat string muodossa.
	 * @return		lauseke, jonka arvon voi laskea
	 * 
	 * @throws UndefinedSymbolException
	 * @throws NotDefinedOperationException
	 * @throws ParseException
	 * @throws NumberFormatException
	 */
	public static Expression Construct(List<String> atoms)
			throws UndefinedSymbolException,
			NotDefinedOperationException,
			ParseException,
			NumberFormatException {
		
		ListIterator<String> i = atoms.listIterator();
		Stack<String> opStack = new Stack<String>();		// operaatiopino
		Stack<Expression> valStack = new Stack<Expression>();	// arvolausekepino
		boolean prevWasOp = true;		// oliko edellinen operaatio (+,-,...,sin,cos,...)
		if( !i.hasNext() )	// tyhj lauseke
			return null;
		String atom;
		do {	// toistetaan niin kauan kuin atomeita riitt
			atom = i.next();
			if( prevWasOp ) {	// edellinen oli thn kohdistuva operaatio
				// sulut
				if( atom.equals("(") ) {
					int begin = i.nextIndex();
					// etsitn loppusulku
					int depth = 1;
					while( depth >= 1 && i.hasNext()) {
						atom = i.next();
						if( atom.equals("(") )
							++depth;
						if( atom.equals(")") )
							--depth;
					}
					if( depth != 0 )	// sulut eptasapainossa, lopetussulkua ei lydy
						throw new ParseException("Sulut eivt ole tasapainossa." +
								(depth == 1 ? " Yksi lopetussulku puuttuu." :
									" " + depth + " lopetussulkua puuttuu."));
					int end = i.previousIndex();
					// rakennetaan alilauseke ja listn pinoon
					valStack.push(Construct(atoms.subList(begin, end)));
					prevWasOp = false;
				// lukuvakio (ensimminen merkki on numero)
				} else if( Character.isDigit(atom.charAt(0)) || atom.charAt(0) == '.' ) {
					try {
						valStack.push(new Constant(Double.parseDouble(atom)));
					} catch( NumberFormatException e) {	// ei sittenkn validi luku
						throw new NumberFormatException("" + atom);
					}
					prevWasOp = false;
				// negaatio
				} else if( atom.equals("-") ) {
					// vaihdetaan merkki erotukseksi miinuksesta
					opStack.push("~");
				// positio
				} else if( atom.equals("+") ) {
					// ei tehd mitn :)
				// funktio
				} else if( functions.contains(atom) ) {
					opStack.push(atom);
				// muuttuja tai literaalivakio
				} else{
					if( !symbols.containsKey(atom) )	// jos ei mritelty => poikkeus
						throw new UndefinedSymbolException("" + atom);
					valStack.push(symbols.get(atom));
					prevWasOp = false;
				}
			} else {	// edellinen oli arvolauseke
				String opNotAllowed;	// alemman prioriteetin operaatiot, joihin lopetetaan laskenta
				boolean funcAllowed;
				
				if( atom.equals("#") )	// mys # toimii potenssina
					atom = "^";
				
				if( atom.equals("+") || atom.equals("-") ) {	// summa tai erotus
					opNotAllowed = "";
					funcAllowed = true;
				} else if( atom.equals("*") || atom.equals("/") ) {	// tulo tai osamr
					opNotAllowed = "+-";
					funcAllowed = true;
				} else if( atom.equals("^") ) {	// potenssi
					opNotAllowed = "+-*/~";
					funcAllowed = false;
				} else if( atom.equals(")") ) {	// ylimrinen lopetussulku
					throw new ParseException("Sulut eivt ole tasapainossa." +
							" Lytyi ylimrinen lopetussulku.");
				} else	// tuntematon operaatio tai operaatiota ei ole
					throw new NotDefinedOperationException("" + atom);
				
				// lasketaan aikaisemmat operaatiot
				calcStack(opStack, valStack, opNotAllowed, funcAllowed);
				opStack.push(atom);	// nykyinen operaatio pinoon
				prevWasOp = true;
			}
		} while( i.hasNext() );
		
		calcStack(opStack, valStack, "", true);	// lasketaan jljell olevat operaatiot
		
		//if( valStack.empty() )
		//	return null;
		return valStack.pop();
	}
	
	/**
	 * Apufunktio, joka kokoaa pinossa olevia operaatioita, kunnes joko operaatiot loppuvat
	 * tai kohdataan liian alhaisen prioriteetin operaatio (nm on annettu parametrina).
	 * @param opStack		operaatiopino
	 * @param valStack		operandipino. Tst otetaan operandit operaatioille, ja thn
	 * 						tynnetn koottu lopputulos.
	 * @param opNotAllowed	operaatiot, joihin funktion suoritus loppuu
	 * @param funcAllowed	kootaanko funktiot (vai pysytetnk niihin)
	 */
	private static void calcStack(Stack<String> opStack, Stack<Expression> valStack,
			String opNotAllowed, boolean funcAllowed) {
		String op;
		Expression lastArg;
		// kootaan operaatioita kunnes pino tyhj tai ei-sallittu operaatio tulee vastaan
		while( !opStack.empty() &&
				((("+-*/^".indexOf(opStack.peek()) >= 0) &&			// (operaatio
					(opNotAllowed.indexOf(opStack.peek()) < 0)) ||	//   ja sallittu) tai
				(("+-*/^".indexOf(opStack.peek()) < 0) &&			// (funktio
					funcAllowed))) {								//   ja sallittu)
			op = opStack.pop();
			lastArg = valStack.pop();
			if( op.equals("+") )		// summa
				valStack.push(new Add(valStack.pop(), lastArg));
			else if( op.equals("-") )	// erotus
				valStack.push(new Sub(valStack.pop(), lastArg));
			else if( op.equals("*") )	// tulo
				valStack.push(new Mul(valStack.pop(), lastArg));
			else if( op.equals("/") )	// osamr
				valStack.push(new Div(valStack.pop(), lastArg));
			else if( op.equals("^") )	// potenssi
				valStack.push(new Pow(valStack.pop(), lastArg));
			else	// ei binrinen operaatio vaan unaarinen funktio
				if( op.equals("~") )		// negaatio
					valStack.push(new Negation(lastArg));
				else if( op.equals("sqrt") )		// nelijuuri
					valStack.push(new Sqrt(lastArg));
				else if( op.equals("cbrt") )		// kuutiojuuri
					valStack.push(new Cbrt(lastArg));
				else if( op.equals("ln") )			// luonnollinen logaritmi
					valStack.push(new Ln(lastArg));
				else if( op.equals("lg") )			// 10-kantainen logaritmi
					valStack.push(new Lg(lastArg));
				else if( op.equals("exp") )			// e^argumentti
					valStack.push(new Exp(lastArg));
				else if( op.equals("abs") )			// itseisarvo
					valStack.push(new Abs(lastArg));
				else if( op.equals("sin") )			// sini
					valStack.push(new Sin(lastArg));
				else if( op.equals("cos") )			// kosini
					valStack.push(new Cos(lastArg));
				else if( op.equals("tan") )			// tangentti
					valStack.push(new Tan(lastArg));
				else if( op.equals("asin") )		// arkussini
					valStack.push(new Asin(lastArg));
				else if( op.equals("acos") )		// arkuskosini
					valStack.push(new Acos(lastArg));
				else if( op.equals("atan") )		// arkustangentti
					valStack.push(new Atan(lastArg));
				else if( op.equals("sinh") )		// hyperbolinen sini
					valStack.push(new Sinh(lastArg));
				else if( op.equals("cosh") )		// hyperbolinen kosini
					valStack.push(new Cosh(lastArg));
				else if( op.equals("tanh") )		// hyperbolinen tangentti
					valStack.push(new Tanh(lastArg));
				else		// funktiota ei tunnistettu (thn ei pitisi pst koskaan)
					throw new ParseException("Tunnistamaton funktio: " + op);
		}
	}
	
	
	
		
	/**
	 * Luo rakentaa parametrina annetusta merkkijonosta muotoa y=f(x) olevan funktion.
	 * @param s		funktion oikean puolen string-esitys
	 * @throws UndefinedSymbolException
	 * @throws NotDefinedOperationException
	 * @throws ParseException
	 * @throws NumberFormatException
	 */
	public FuncXtoY(String s)
			throws UndefinedSymbolException,
			NotDefinedOperationException,
			ParseException,
			NumberFormatException {
		this.rightSide = Construct(Split(s));
	}
	
	/**
	 * Laskee funktion arvon parametrina annetussa kohdassa.
	 * @param x		parametri x
	 * @return	funktion arvo
	 */
	public double calcValue(double x) {
		varX.setValue(x);
		return this.rightSide.getValue();
	}
	
	//----------------------------------------------------------------------------------
	// Testaukseen tarkoitetut metodit:
	
	/**
	 * Testauksessa hydynnettv funktio, joka tarkastaa, ovatko liukuluvut kytnnss
	 * samoja.
	 * @param val1,val2	testattavat arvot
	 */
	private static boolean close(double val1, double val2) {
		return Math.abs(val2 / val1 - 1.0) < .001;
	}
	
	/**
	 * Testausapufunktio, joka tulostaa annetun testin string-esityksen, sek tiedon testin
	 * onnistumisesta ruudulle. Tuloste pilkkoutuu alle 80 merkin mittaisiksi riveiksi.
	 * @param s
	 * @param succ
	 */
	private static void printResult(String s, boolean succ) {
		while( s.length() > 70 ) {
			System.out.print(s.substring(0, 70) + "..\n");
			s = ".." + s.substring(70);
		}
		char[] space = new char[75 - s.length()];
		java.util.Arrays.fill(space, ' ');
		s += new String(space);
		System.out.println( s + (succ ? "OK" : "FAIL!"));	}
	
	/**
	 * Testausfunktio tutkii, tsmk lausekkeen arvo annettuun vertailuarvoon, ja 
	 * tulostaa verrattavan lausekkeen, tulosarvon sek tsmsik se vertailuarvoon.
	 * Tulostaa tuloksen nytlle.
	 * @param e		lauseke, jonka arvoa tutkitaan
	 * @param val	vertailuarvo
	 */
	private static void testValue(Expression e, double val) {
		String s = e + " = " + e.getValue() + " (" + val + ")";
		printResult(s, close(e.getValue(), val));
	}
	
	/**
	 * Lausekkeen sisvvn string-muuttujan paloittelun testaukseen kytettv metodi.
	 * Tulostaa alkuperisen lausekkeen, sek paloitellun version palat vlilynnill
	 * eroteltuina.
	 * @param s		paloiteltava merkkijono
	 */
	private static void testSplit(String s) {
		List<String> atoms = Split(s);
		System.out.println(s + "\n => ");
		for( int i = 0; i < atoms.size(); ++i )
			System.out.print(" " + atoms.get(i));
		System.out.println();
	}
	
	/**
	 * Testausfunktio, joka tulostaa parametrina annetun merkkijonon, ja siit johdetun
	 * kaavan kyttjn vertailtavaksi.
	 * @param s		vertailtava kaava
	 */
	private static void printConstructTest(String s) {
		System.out.println(s + "\n => " + Construct(Split(s)));
	}
	
	/**
	 * FuncXtoY-luokan testaukseen tarkoitettu apumetodi. Luo ko. luokan ilmentymn
	 * parametrina annetusta lausekkeesta. Laskee lausekkeen arvon annetussa pisteess,
	 * ja vertaa sit valmiiksi annettuun vertailuarvoon. Tulostaa lausekkeen, lasketun
	 * arvon sek vertailun tuloksen nytlle.
	 * @param expr	lauseke, josta funktion generoidaan
	 * @param x		piste, jossa funktion arvo lasketaan
	 * @param val	vertailuarvo
	 */
	private static void testFuncXtoY(String expr, double x, double val) {
		FuncXtoY f = new FuncXtoY(expr);
		String s = "y = " + f.rightSide +  " = " + f.calcValue(x) + " (" + val + ")";
		printResult(s, close(f.calcValue(x), val));
	}
	
	/**
	 * Testaukseen tarkoitettu main-metodi.
	 * @param args
	 */
	public static void main(String[] args) {
		//Scanner cin = new java.util.Scanner(System.in);
		
		//String kaava = cin.nextLine();
		//double val1, val2;
		
		//Expression e;
		
		// operaatioiden testaus
		System.out.println("Operaatioiden testaus");
		testValue(new Add(CONST_PI, CONST_PI), Math.PI + Math.PI);
		testValue(new Sub(new Constant(1.0), new Constant(-1.0)), 2.0);
		testValue(new Mul(new Constant(2.0), new Constant(-1.5)), -3.0);
		testValue(new Div(new Constant(2.0), new Constant(-1.5)), -2.0/1.5);
		testValue(new Pow(new Constant(-2.0), new Constant(-1.0)), -.5);
		
		// funktioiden testaus
		System.out.println("\nFunktioiden testaus");
		testValue(new Sin(new Constant(-5)), Math.sin(-5));
		testValue(new Cos(new Constant(-5)), Math.cos(-5));
		testValue(new Tan(new Constant(-5)), Math.tan(-5));
		testValue(new Tan(new Constant(Math.PI/2)), Math.tan(Math.PI/2));
		testValue(new Asin(new Constant(0.1)), Math.asin(0.1));
		testValue(new Acos(new Constant(-1)), Math.acos(-1));
		testValue(new Atan(new Constant(-5)), Math.atan(-5));
		testValue(new Sinh(new Constant(-5)), Math.sinh(-5));
		testValue(new Cosh(new Constant(-5)), Math.cosh(-5));
		testValue(new Tanh(new Constant(-5)), Math.tanh(-5));
		testValue(new Ln(new Constant(100)), Math.log(100));
		testValue(new Lg(new Constant(30)), Math.log10(30));
		testValue(new Exp(new Constant(-5)), Math.exp(-5));
		testValue(new Sqrt(new Constant(30)), Math.sqrt(30));
		testValue(new Cbrt(new Constant(-5)), Math.cbrt(-5));
		testValue(new Abs(new Constant(-5)), Math.abs(-5));
		
		// Testataan kaavanpilkkojaa
		System.out.println("\nTestataan kaavanpilkkojaa");
		testSplit("1+.2-3E1*-7.406/.01E-4.2^0E++6");
		testSplit(")x+)+-  (n2sin xxx /*+- 2--)^ln -abs help2 cos (((x + sqrt x*-.3E-2)");
		
		// Testataan kaavantulkintaa kokonaisuudessaan
		System.out.println("\nTestataan kaavantulkintaa kokonaisuudessaan");
		printConstructTest("sin -2");
		printConstructTest("1 + exp 5 ^ 2 * 3");
		printConstructTest("1 + exp 5 ^ 2 ^ 3");
		printConstructTest("cos -2^3");
		printConstructTest("(cos -2)^3");
		printConstructTest("1^sin 2^sin 3^sin 4");
		printConstructTest("-2+3");
		printConstructTest("-2*3");
		printConstructTest("-2/3");
		printConstructTest("-2^3");
		printConstructTest("1+2*3^4");
		printConstructTest("4^3/2-1");
		printConstructTest("-2/3");
		printConstructTest("0-2/3");
		
		// Testataan koko homman toimintaa
		System.out.println("\nTestataan koko homman toimintaa");
		testFuncXtoY("x+1", 5, 6);
		testFuncXtoY("3*x^2", 4, 48);		
		testFuncXtoY("2*x^4 +  x^2/2 + .5*x^-1 + 1", 2, 35.25);
		testFuncXtoY("sin cos x", -1, .514395258);
		testFuncXtoY("x^x^x", 3, 19683);
		testFuncXtoY("x^(x^x)", 3, 7.625597485e12);
		
//		varX.setValue(1);
//		System.out.println(e.getValue());
//		varX.setValue(2);
//		System.out.println(e.getValue());
		
//		kaava = "x";
//		FuncXtoY f = new FuncXtoY(kaava);
//		System.out.println(f.calcValue(2));
	}
}

