package funktiopiirturi;

import javax.swing.*;
import java.awt.*;
//import java.awt.event.*;
import java.util.Vector;

/**
 * XY-koordinaatistoa/piirtopintaa esittv luokka. On oikeastaan nkym/ikkuna sellaiseen.
 * Sislt piirrettvt funktiot.
 *
 */
class SurfaceXY extends JComponent {
	final static long serialVersionUID = 1L;				// varoitukset pois...
	private static final int PIXELS_PER_ONE = 64;			// yksikn koko perustilassa
	//private static final int LINE_STEP_SHOW_EVERY_SECOND = 64;
	private static final String ABS_EXP_SHOWN_NORMALLY = "000";	// max nollia luvuissa
	
	//private static final double PIXEL_SIZE = 1.0 / PIXELS_PER_ONE; //.03125;;
	private Vector<FuncXtoY> functions;				// kaikki piirrettvt funktiot
	private double zoomX, zoomY;					// suurennoskertoimet eri akseleille
	private double posX, posY;						// tarkasteltava koordinaatiston kohta
	private int imgCenterX, imgCenterY;				// kuvan keskipiste
	private double left, right, top, bottom;		// nkymn reunat
	private double pixelsPerOneX, pixelsPerOneY; 	// yhden yksikn koko pikselein
	private double pixelWidth, pixelHeight;			// pikselin koko (yksikk)
	
	//koordinaattiviivoihin liittyv tavaraa
	private static final int MIN_LINE_STEP_PIXELS = 32;		// min koordinaatistoviivavli
	private double coordLineStepX, coordLineStepY;			// viivojen vlit eri suunnissa
	private long coordLineExpX, coordLineExpY;				// log10 edellinen				
	private long firstHCoordLineNum, lastHCoordLineNum;		// ekat ja vikat viivat  "
	private long firstVCoordLineNum, lastVCoordLineNum;		//
	
	/**
	 * Luo tyhjn XY-koordinaatiston, ja asettaa sen perustilaan.
	 */
	public SurfaceXY() {
		this.functions = new Vector<FuncXtoY>();
		this.zoomX = 1.0;
		this.zoomY = 1.0;
		this.posX = 0.0;
		this.posY = 0.0;
		this.reCalculate();
	}
	
	/**
	 * Muuttaa kuvan kokoa ja/tai aselmointia, ja laskee tarvittavat arvot uudelleen.
	 */
	public void setBounds(int x, int y, int width, int height) {
		super.setBounds(x, y, width, height);
		this.reCalculate();
	}
	
	/**
	 * Piirt koordinaatiston ja kuvaajat.
	 * @param		g	kuvapinta, jolle piirtminen tapahtuu
	 */
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		
		// valkoinen tausta
		g.setColor(Color.WHITE);
        g.fillRect(0, 0, getWidth(), getHeight());
		
		// piirretn koordinaatisto
		this.drawCoords(g);
		
		// piirretn kuvaajat
		g.setColor(Color.BLACK);	// mustalla vrill
		for( int i = 0; i < this.functions.size(); ++i )
			this.drawFunc(g, this.functions.get(i));
	}
	
	/**
	 * Piirt annetun funktion. Piirtovrin kytetn viimeksi asetettua. Huom. tm ei
	 * toimi idioottivarmasti; Jotkin funktioiden epjatkuvuuskohdat saattavat piirty
	 * vrin.
	 * @param	g	grafiikkapinta, jolle piirtminen tapahtuu
	 * @param	f	piirrettv funktio
	 */
	private void drawFunc(Graphics g, FuncXtoY f) {
		double x = this.left;
		double imgYLeft, imgYRight;	// kuvaajan y-koordinaatit pikselisarakkeen laidoilla
		imgYRight = (int)((this.top - f.calcValue(x)) *
				this.pixelsPerOneY);
		
		// kydn lpi jokainen pikselisarake vasemmalta oikealle
		for( int imgX = 0; imgX < this.getWidth(); ++imgX ) {
			imgYLeft = imgYRight;
			x = this.left + (imgX + 1) * this.pixelWidth;	// sarakkeen oikean laidan x
			imgYRight = ((this.top - f.calcValue(x)) *
					this.pixelsPerOneY);
			
			// mrittelemttmi tuloksia ei piirret
			if( Double.isNaN(imgYLeft) || Double.isNaN(imgYRight) )
				continue;
			
			double yTop = Math.min(imgYLeft, imgYRight);	// ylempi sarakkeen laidoista
			double yBottom = Math.max(imgYLeft, imgYRight);	// alempi
			
			// jos kokonaan kuvan ulkopuolella, niin ei piirret
			if( yBottom < 0.0 || yTop > this.getHeight() )
				continue;
			
			// suljetaan pois rettmst toiseen hyppvien funktioiden epjatkuvuuskohdat
			// kaipaa parempaa tarkistusta! (ei toimi varmasti, ja sulkee pois muitakin)
			if( (yTop < -this.getHeight()*1) && (yBottom > this.getHeight()*2) )
				continue;
			
			// y-arvo pikselisarakkeen keskell
			double yMid = (yTop + yBottom) / 2;
			
			// piirrettn sarake
			yTop += .5;
			yBottom -= .5;
			if( yTop < yBottom )
				g.drawLine(imgX, (int)yTop, imgX, (int)yBottom);
			
			// silt varalta, ett korkeus oli alle 1
			g.drawLine(imgX, (int)yMid, imgX, (int)yMid);
		}
	}
	
	/**
	 * Piirt koordinaatiston. Tm metodi piirt valitsemansa kymmenen potenssin vlein
	 * koordinaattiviivoja, ja merkitsee niiden etisyyden origosta. Mahdollisesti kuvassa
	 * nkyvt koordinaattiakselit piirretn hieman tummemmalla.
	 * @param	g	kuvapinta, jolle piirtminen tapahtuu
	 */
	private void drawCoords(Graphics g) {
		// ensi piirretn vaakaviivat
		for( long i = this.firstHCoordLineNum; i <= this.lastHCoordLineNum; ++i ) {
			if( i == 0 )	// asetetaan vri (akselit tummemmalla)
				g.setColor(Color.DARK_GRAY);
			else
				g.setColor(Color.LIGHT_GRAY);
			
			// lasketaan kohta ruudulla
			double y = i * this.coordLineStepY;
			int imgY = (int)((this.top - y) * this.pixelsPerOneY);
			g.drawLine(0, imgY, this.getWidth()-1, imgY);				// piirretn viiva
			g.setColor(Color.GRAY);
			g.drawString(toString(i, this.coordLineExpY), 0, imgY-1);	// etisyys origosta
		}
		
		// sitten pystyviivat vastaavasti
		for( long i = this.firstVCoordLineNum; i <= this.lastVCoordLineNum; ++i ) {
			if( i == 0 )
				g.setColor(Color.DARK_GRAY);
			else
				g.setColor(Color.LIGHT_GRAY);
			double x = i * this.coordLineStepX;
			int imgX = (int)((x - this.left) * this.pixelsPerOneX);
			g.drawLine(imgX, 0, imgX, this.getHeight()-1);
			g.setColor(Color.GRAY);
			g.drawString(toString(i, this.coordLineExpX), imgX+1, g.getFont().getSize());
		}
	}
	
	/**
	 * Laskee funktioiden ja koordinaatiston piirtmisess tarvittavat apuarvot.
	 * Tt metodia kutsutaan kun esimerkiksi koordinaatiston kokoa tai suurennoskerrointa
	 * muutetaan.
	 */
	private void reCalculate() {
		// tarkemmat selitykset muuttujista lytyvt niiden mrittelyn yhteydest
		
		// kuvan ominaisuudet
		this.imgCenterX = this.getSize().width / 2;
		this.imgCenterY = this.getSize().height / 2;
		this.pixelsPerOneX = PIXELS_PER_ONE * this.zoomX;
		this.pixelsPerOneY = PIXELS_PER_ONE * this.zoomY;
		this.pixelWidth = 1.0 / this.pixelsPerOneX;
		this.pixelHeight = 1.0 / this.pixelsPerOneY;
		this.left = this.posX - (this.imgCenterX + .5) * pixelWidth;
		this.right = this.posX + ((this.getWidth()-this.imgCenterX) - .5) * pixelWidth;
		this.top = this.posY + (this.imgCenterY + .5) * pixelHeight;
		this.bottom = this.posY - ((this.getHeight()-this.imgCenterY) - .5) * pixelHeight;
		
		// koordinaatistoviivat
		this.coordLineExpX = Math.round(Math.ceil(
				Math.log10(MIN_LINE_STEP_PIXELS * this.pixelWidth)));
		this.coordLineExpY = Math.round(Math.ceil(
				Math.log10(MIN_LINE_STEP_PIXELS * this.pixelHeight)));
		this.coordLineStepX = Math.pow(10, coordLineExpX);
		this.coordLineStepY = Math.pow(10, coordLineExpY);
		this.firstHCoordLineNum = Math.round(Math.ceil(this.bottom / this.coordLineStepY));
		this.lastHCoordLineNum = Math.round(Math.floor(this.top / this.coordLineStepY));
		this.firstVCoordLineNum = Math.round(Math.ceil(this.left / this.coordLineStepX));
		this.lastVCoordLineNum = Math.round(Math.floor(this.right / this.coordLineStepX));
	}
	
	/**
	 * Lis funktion piirrettvien funktioiden listaan. Aiheuttaa kuvan piirtmisen
	 * uudelleen.
	 * @param	f	listtv funktio
	 */
	public void addFunction(FuncXtoY f) {
		if( f != null ) {
			this.functions.add(f);
			this.repaint();
		}
	}
	
	/**
	 * Asettaa horisontaalisen suurennoskertoimen, ja aiheuttaa kuvan piirtmisen uudestaan.
	 * Arvot jotka ovat pienempi tai yhtsuuria kuin nolla eivt muuta suurennosta. 
	 * @param newZoomX	uusi suurennoskerroin
	 */
	public void setZoomX(double newZoomX) {
		if( newZoomX > 0.0 ) {
			this.zoomX = newZoomX;
			this.reCalculate();
			this.repaint();
		}
	}
	
	/**
	 * Asettaa vertikaalisen suurennoskertoimen, ja aiheuttaa kuvan piirtmisen uudestaan.
	 * Arvot jotka ovat pienempi tai yhtsuuria kuin nolla eivt muuta suurennosta. 
	 * @param newZoomY	uusi suurennoskerroin
	 */
	public void setZoomY(double newZoomY) {
		if( newZoomY > 0.0 ) {
			this.zoomY = newZoomY;
			this.reCalculate();
			this.repaint();
		}
	}
	
	/**
	 * Palauttaa tmnhetkisen horisontaalisen suurennoskertoimen.
	 * @return	suurennoskerroin
	 */
	public double getZoomX() {
		return this.zoomX;
	}
	
	/**
	 * Palauttaa tmnhetkisen vertikaalisen suurennoskertoimen.
	 * @return	suurennoskerroin
	 */
	public double getZoomY() {
		return this.zoomY;
	}
	
	/**
	 * Poistaa kaikki funktiot ja tyhjent piirtopinnan.
	 */
	public void clear() {
		this.functions.clear();
		this.repaint();
	}
	
	/**
	 * Apufunktio, jota kytetn koordinaattiviivojen merkitsemiseen. Palauttaa mantissan
	 * ja eksponentin perusteella oikean luvun halutusti muotoillun string-esityksen.
	 * @param mant	luvun mantissa
	 * @param exp	luvun eksponentti
	 * @return	string-esitys luvusta
	 */
	private static String toString(long mant, long exp) {
		if( mant == 0 )		// nolla palautetaan nollana
			return "0";
		if( Math.abs(exp) <= ABS_EXP_SHOWN_NORMALLY.length() ) {	// normaali muoto
			if( exp < 0 )
				return (mant < 0 ? "-" : "") + "0." + ABS_EXP_SHOWN_NORMALLY.substring(
						(int)(ABS_EXP_SHOWN_NORMALLY.length()+exp+1)) + Math.abs(mant);
			else
				return mant + ABS_EXP_SHOWN_NORMALLY.substring(
						(int)(ABS_EXP_SHOWN_NORMALLY.length()-exp));
		} else
			return mant + "e" + exp;	// tieteellinen muoto
	}
	
	/**
	 * Testausmetodi. Piirtelee muutamia kuvaajia ruudulle.
	 * @param args	ohjelmalle vlitettvt parametrit (ei vaikutusta)
	 */
	public static void main(String[] args) {
		JFrame frame = new JFrame("Testaus");
		SurfaceXY s = new SurfaceXY();
		frame.getContentPane().add(s);
		frame.setPreferredSize(new Dimension(600, 600));
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.pack();
		frame.setVisible(true);
		s.addFunction(new FuncXtoY("x"));
		s.addFunction(new FuncXtoY("2*x"));
		s.addFunction(new FuncXtoY("10*x"));
		s.addFunction(new FuncXtoY(".5*x"));
		s.addFunction(new FuncXtoY(".1*x"));
		s.addFunction(new FuncXtoY("-x"));
		s.addFunction(new FuncXtoY("x^2-x-1"));
		s.addFunction(new FuncXtoY("tan x"));
		s.addFunction(new FuncXtoY("sqrt x"));
		s.addFunction(new FuncXtoY("ln x"));
		s.addFunction(new FuncXtoY("tan x"));
	}
}
